"""General purpose utilities."""

from colorama import Fore
from datetime import datetime, time
from functools import update_wrapper, partial, wraps
from inspect import _empty, Parameter, signature
import json
import logging
import os
from pathlib import Path
import re
import sys
from threading import Thread
import warnings
import yaml
import _thread

from htools import select, bound_args, copy_func, add_docstring, \
    listlike, MultiLogger, tolist, load, save
from jabberwocky.config import C


class ReturningThread(Thread):

    @add_docstring(Thread)
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, *, daemon=None):
        """This is identical to a regular thread except that the join method
        returns the value returned by your target function. The
        Thread.__init__ docstring is shown below for the sake of convenience.
        """
        super().__init__(group=group, target=target, name=name,
                         args=args, kwargs=kwargs, daemon=daemon)
        self.result = None

    def run(self):
        self.result = self._target(*self._args, **self._kwargs)

    def join(self, timeout=None):
        super().join(timeout)
        return self.result


def interrupt_on_complete(meth):
    """Decorator that powers PropagatingThread. We do this here rather than
    interrupting directly from run() method because run must return in order
    for thread to stop, but of course we can't do anything directly from the
    method after returning.
    """
    @wraps(meth)
    def wrapper(*args, **kwargs):
        res = meth(*args, **kwargs)
        if args[0].exception and args[0].raise_immediately:
            print('INTERRUPTING')
            _thread.interrupt_main()
        return res
    return wrapper


class PropagatingThread(Thread):
    """Ported from gui/utils.py since I realized ReturningThread never raises
    error to calling context if it fails.

    Thread that will raise an exception in the calling thread as soon as one
    occurs in the worker thread. You must use a KeyboardInterrupt or
    BaseException in your try/except block since that is the only type of
    exception we can raise. If you don't need the exception to be raised
    immediately, you can set raise_immediately=False and it won't be propagated
    until the thread is joined.

    thread.join() also returns the value returned by the thread's target
    function.

    Partially based on version here, though that can only raise on join():
    https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread
    """

    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, *, daemon=None,
                 raise_immediately=False):
        """
        Parameters
        ----------
        raise_immediately: bool
            If True, raise any exception as soon as it occurs. In order to
            propagate this to the calling thread, we are forced to use a
            KeyboardInterrupt. No information about the actual exception will
            be propagated. If False, the actual exception will be raised on
            the call to thread.join().
        """
        super().__init__(group=group, target=target, name=name,
                         args=args, kwargs=kwargs, daemon=daemon)
        self.raise_immediately = raise_immediately
        self.exception = None
        self.result = None

    @interrupt_on_complete
    def run(self):
        try:
            self.result = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e

    def join(self, timeout=None):
        super().join(timeout)
        if not self.raise_immediately and self.exception:
            raise self.exception
        return self.result


def thread_starmap(func, kwargs_list=None, raise_errors=True,
                   raise_immediately=False):
    """Similar to multiprocessing.Pool.starmap but with my ReturningThread
    instead.

    Parameters
    ----------
    raise_errors: bool
        If True, any error in one of the underlying threads will result in an
        error being raised to the calling context. If False, no error will be
        raised and the corresponding list item will simply be None.
    raise_immediately: bool
        If raise_errors is False, this is unused. If it's True, it determines
        whether exceptions that occur in threads are raised immediately when
        they happen (if True) or only when the failed thread is joined
        (if False).
    """
    kwargs_list = kwargs_list or [{}]
    if raise_errors:
        thread_cls = partial(PropagatingThread,
                             raise_immediately=raise_immediately)
    else:
        thread_cls = ReturningThread
    threads = [thread_cls(target=func, kwargs=kwargs)
               for kwargs in tolist(kwargs_list)]
    for thread in threads: thread.start()
    return [thread.join() for thread in threads]


class JsonlinesFormatter(logging.Formatter):
    """Formatter for logging python data structures to a jsonlines file.
    Used by JsonLogger.
    """

    def format(self, record):
        return json.dumps(record.msg)


class JsonlinesLogger(MultiLogger):
    fmt = '%(message)s'

    def __init__(self, path):
        super().__init__(path, fmode='a', fmt=self.fmt)
        self.formatter = self._add_json_formatter()
        self.path = self._get_file_handler()[0].baseFilename

    def _log(self, level, msg, args, exc_info=None, extra=None,
             stack_info=False):
        if not os.path.isfile(self.path):
            raise RuntimeError('PATH DOES NOT EXIST.')
        super()._log(level, msg, args, exc_info=exc_info, extra=extra,
                     stack_info=stack_info)

    def _get_file_handler(self):
        """
        Returns
        -------
        tuple: First item is FileHandler, second is its index in the logger's
        file handlers. This allows us to effectively edit it.
        """
        handlers = [(handler, i) for i, handler in enumerate(self.handlers)
                    if isinstance(handler, logging.FileHandler)]
        if len(handlers) != 1:
            raise RuntimeError(
                'Expected JsonlinesLogger to have 1 FileHandler, '
                f'found {len(handlers)}.'
            )
        return handlers[0]

    def _add_json_formatter(self):
        formatter = JsonlinesFormatter(self.fmt)
        handler, _ = self._get_file_handler()
        handler.setFormatter(formatter)
        return formatter

    def change_path(self, path):
        """Change the path of the file we log to by creating a new FileHandler
        because the filepath affects other attributes that are set in the
        constructor.
        """
        _, i = self._get_file_handler()
        self.handlers[i] = self._change_path(self.handlers[i], path)
        self.path = self.handlers[i].baseFilename

    def _change_path(self, handler, path):
        kwargs = {key: getattr(handler, key) for key in
                  ('mode', 'encoding', 'delay')}
        handler = type(handler)(filename=path, **kwargs)
        handler.setFormatter(self.formatter)
        return handler


def with_signature(to_f, keep=False):
    """Decorator borrowed from fastai and renamed to avoid name collision
    with htools (originally called "delegates"). Replaces `**kwargs`
    in signature with params from `to`. Unlike htools.delegates, it only
    changes documentation - variables are still made available in the decorated
    function as 'kwargs'.
    """
    def _f(f):
        from_f = f
        sig = signature(from_f)
        sigd = dict(sig.parameters)
        k = sigd.pop('kwargs')
        s2 = {k: v for k, v in signature(to_f).parameters.items()
              if v.default != Parameter.empty and k not in sigd}
        sigd.update(s2)
        if keep: sigd['kwargs'] = k
        from_f.__signature__ = sig.replace(parameters=sigd.values())
        return f
    return _f


def strip(text, do_strip=True):
    """Convenience function used in query_gpt3.

    Parameters
    ----------
    text: str
    do_strip: bool

    Returns
    -------
    str: If do_strip is False, this is the same as the input.
    """
    return text.strip() if do_strip else text


def squeeze(*args, n=1):
    """Return either the input `args` or the first item of each arg, depending
    on our choice of n. We effectively treat n as a boolean. This is used by
    the query_gpt_{} functions to return either a (str, dict) tuple or a
    (list[str], list[dict]) tuple, depending on the number of completions
    we ask for. If first arg is not some kind of list/tuple (e.g. a str), we
    simply return the args unchanged.
    """
    if not listlike(args[0]):
        return args
    return tuple(arg[0] for arg in args) if n == 1 else args


def touch(path):
    """Create an empty file, like bash `touch` command. If necessary, we
    create the parent dir as well.
    """
    os.makedirs(Path(path).parent, exist_ok=True)
    with open(path, 'a'):
        pass


def load_booste_api_key():
    """Load api key for booste, a way to programmatically access gpt2 and clip.

    Returns
    -------
    str
    """
    raise DeprecationWarning('Booste.ai no longer exists. It was resurrected '
                             '(or at least rebranded) as free.banana.dev. '
                             'Try using `load_banana_api_key`.')
    return load_api_key('booste')


def load_huggingface_api_key():
    """Load api used when querying Huggingface model APIs. This will be placed
    in the headers for a post request like
    {'Authorization': f'Bearer api_{my_api_key}'}.
    """
    return load_api_key('huggingface')


def load_api_key(name, raise_error=True):
    """Generic api key loader. Assumes you store the key as a text file
    containing only one key in a file called ~/.{name}.

    Parameters
    ----------
    name: str
        Examples: 'booste', 'huggingface', 'gooseai'.
    raise_error: bool
        If True, a missing file will raise an error. If False, a missing file
        will only cause a warning and will return an empty string.

    Returns
    -------
    str
    """
    def _load_key(name):
        """Load api key or raise FileNotFoundError. We also check for
        an environment variable like OPENAI_API_KEY.
        """
        try:
            with open(Path(f'~/.{name}').expanduser(), 'r') as f:
                return f.read().strip()
        except FileNotFoundError as e:
            key = os.getenv(f'{name.upper()}_API_KEY')
            if key: return key
            raise e

    if raise_error:
        return _load_key(name)
    try:
        return _load_key(name)
    except FileNotFoundError:
        warnings.warn(
            f'Jabberwocky expected to find an api key for backend '
            f'{name} but couldn\'t. Using {name} backend will throw '
            f'auth errors. Add your api key to "~/.{name} (if you are using a '
            f'GPTBackend object, you will then need to use its '
            f'refresh_api_keys method to enable the backend.'
        )
        return ''


load_goose_api_key = partial(load_api_key, name='gooseai')


def save_yaml(data, path):
    """Save a dictionary as a yaml file.

    Parameters
    ----------
    data: dict
    path: str or Path
        Any intermediate dirs that do not already exist will be created.
    """
    path = Path(path)
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, 'w') as f:
        yaml.dump(data, f)


def load_yaml(path, section=None):
    """Load a yaml file. Useful for loading prompts.

    Parameters
    ----------
    path: str or Path
    section: str or None
        I vaguely recall yaml files can define different subsections. This lets
        you return a specific one if you want. Usually leave as None which
        returns the whole contents.

    Returns
    -------
    dict
    """
    with open(path, 'r') as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    return data.get(section, data)


def escape_markdown(text):
    raw = r'([!#()*+-.[\]_`{|}])'
    return re.sub(raw, r'\\\1', text)


def _update_prompt_readme(dir_='data/prompts', sep='***'):
    """Update readme file in a directory of prompt yaml files to contain a
    table displaying their name and description (provided by the "doc" field
    in the yaml file.
    """
    dir_ = Path(dir_)
    name2doc = {path.stem: escape_markdown(load_yaml(path).get('doc', '')
                                           .replace('\n', ' '))
                for path in dir_.iterdir() if path.suffix == '.yaml'}
    tbody = '\n'.join(f'{k} | {v}' for k, v in sorted(name2doc.items(),
                                                      key=lambda x: x[0]))
    table = f'Prompt Name | Doc\n---|---\n{tbody}'
    try:
        readme = load(dir_/'README.md')
    except FileNotFoundError:
        readme = sep
    keep, sep, _ = readme.partition(sep)
    readme = f'{keep}{sep}\n{table}'
    save(readme, dir_/'README.md')
    return readme


def bold(text):
    """Make text bold when printed. Note that without the print statement, this
    will just add a weird character string to the front and end of the input.

    Parameters
    ----------
    text: str

    Returns
    -------
    str
    """
    return C.bold_start + text + C.bold_end


def underline(text):
    """Make text underlined when printed.

    Parameters
    ----------
    text: str

    Returns
    -------
    str
    """
    return C.underline_start + text + C.bold_end


def colored(text, color):
    """Update string so it will show up in a different color when printed. Note
    that the raw string itself will now have special characters added to the
    beginning and end, so the repr will look a bit odd.

    Parameters
    ----------
    text: str
    color: str
        Must be a color provided by colorama.Fore.

    Returns
    -------
    str
    """
    return getattr(Fore, color.upper()) + text + Fore.RESET


def getindex(arr, val, default=-1):
    return arr.index(val) if val in arr else default


def most_recent_filepath(dir_, mode='m'):
    """Get path of most recently modified file in a directory.

    Parameters
    ----------
    dir_: str or Path
        Directory to look in.
    mode: str
       One of ('m', 'c') corresponding to mtime or ctime. ctime changes when
       file permissions change while mtime does not, but they're very similar
       otherwise.

    Returns
    -------

    """
    paths = [path for path in Path(dir_).iterdir() if path.is_file()]
    if not paths:
        raise RuntimeError(f'No files in directory {dir_}.')
    return max(paths, key=lambda x: getattr(x.stat(), f'st_{mode}time'))


def set_module_global(module, key, value):
    """Create global variable in an imported module. This is a slightly hacky
    workaround that solves some types of circular imports.

    Parameters
    ----------
    module: str
    key: str
    value: any
    """
    setattr(sys.modules[module], key, value)


def set_module_globals(module, **kwargs):
    """Set multiple global variables in an imported module."""
    for k, v in kwargs.items():
        set_module_global(module, k, v)


def hooked_generator(gen, *hooks):
    """Attach hook(s) to a generator. Motivation: want to be able to return
    query_gpt3 response in Conversationmanager even if stream=True and still
    ensure that updates are made to the conversation history.

    Parameters
    ----------
    gen
    hooks: Iterable[function]
        1 or more functions that accept three arguments: the first is
        the object yielded by the generator at each step, the second is the
        index (starting at zero) of the current step, and the third is a
        boolean value `is_post` specifying whether we're calling it after the
        generator has finished executing. Any return value is
        ignored - it simply executes. If multiple hooks are provided, they will
        be executed in the order they were passed in.

    Yields
    ------
    Values are yielded from the input generator. They will be unchanged unless
    they are mutable AND one or more hooks alter them.
    """
    for i, val in enumerate(gen):
        for hook in hooks:
            hook(val, i)
        yield val
    for hook in hooks:
        hook(val, i, is_post=True)


class Partial:
    """More powerful (though also potentially more fragile) version of
    functools.partial that updates the resulting signature to work better with
    Jupyter's quick documentation feature. We also update __repr__, __str__,
    and __name__ attributes (optionally renaming the source function). Unlike
    functools.partial, we also reorder parameters so that those without
    defaults always come before those with defaults.

    Note: the resulting object is actually a callable class, not a function.
    """

    def __init__(self, func, name=None, **kwargs):
        """
        Parameters
        ----------
        func: function
        name: str or None
            If None, the source function's name will be used.
        kwargs: any
            Default arguments to set, like in functools.partial.
        """
        self.func = copy_func(func)
        self.old_name = func.__name__

        # Track names of positional args in old function since this affects
        # the order args must be passed in if var_positional parameters
        # (*args) are present.
        self.old_pos_pars = []
        self.kwargs_name = ''
        self.args_name = ''
        new_pars = []
        old_sig = signature(self.func)
        for k, v in old_sig.parameters.items():
            # Check parameter kind for error handling and argument resolution
            # in __call__.
            if v.kind == 0:
                raise NotImplementedError(
                    'rigorous_partial does not support functions with '
                    'positional only parameters.'
                )
            elif v.kind == 2:
                self.args_name = k
            elif v.kind == 4:
                self.kwargs_name = k
                break

            if v.kind <= 2:
                self.old_pos_pars.append(k)

            # Assign default value from newly specified kwargs if provided.
            if k in kwargs:
                default = kwargs.pop(k)
                kind = 3
            else:
                default = v.default
                kind = v.kind
            param = Parameter(k, kind, default=default)
            new_pars.append(param)

        # Remaining kwargs only: those that were not present in func's
        # signature. Require that they be keyword only since ordering can
        # cause issues (updating signature affects what we see but doesn't
        # seem to affect the actual order args are passed in, presumably due
        # to old __code__ object).
        for k, v in kwargs.items():
            param = Parameter(k, 3, default=v)
            new_pars.append(param)
        if self.kwargs_name:
            new_pars.append(Parameter(self.kwargs_name, 4))

        # Ensure we don't accidentally place any parameters with defaults
        # ahead of those without them. Third item in tuple is a tiebreaker
        # (defaults to original function's parameter order).
        old_names = [p for p in old_sig.parameters]
        new_pars.sort(
            key=lambda x: (x.kind, x.default != _empty,
                           getindex(old_names, x.name, float('inf')))
        )

        # I honestly forget why we need to set the attribute on self.func too,
        # I just remember it was needed to resolve a bug (I think it was
        # related to *args resolution).
        self.__signature__ = self.func.__signature__ = old_sig.replace(
            parameters=new_pars
        )
        self.__defaults__ = tuple(p.default for p in new_pars if p.kind < 3
                                  and p.default != _empty)
        self.__kwdefaults__ = {p.name: p.default for p in new_pars
                               if p.kind == 3}
        if name: self.func.__name__ = name
        update_wrapper(self, self.func)

    def __call__(self, *args, **new_kwargs):
        # Remember self.func's actual code is unchanged: we updated how its
        # signature appears, but that doesn't affect the actual mechanics.
        # Therefore, we need to carefully resolve args and kwargs so that the
        # function is called so that behavior matches what we'd expect based
        # on the order shown in the signature.
        tmp_kwargs = bound_args(self.func, args,
                                {**self.__kwdefaults__, **new_kwargs})
        final_args = {name: tmp_kwargs.pop(name)
                      for name in self.old_pos_pars}
        final_star_args = final_args.pop(self.args_name, [])
        final_kwargs = select(tmp_kwargs, drop=list(final_args))
        return self.func(*final_args.values(), *final_star_args,
                         **final_kwargs)

    def __repr__(self):
        """Note: the memory address here points to that of the copy of the
        source function stored in self.func.
        """
        return repr(self.func).replace(self.old_name, self.__name__)

    def __str__(self):
        return str(self.func).replace(self.old_name, self.__name__)


def containerize(*args, dtype=list):
    """Basically applies tolist() to a bunch of objects, except that tolist
    doesn't work on dicts. I considered changing that but I'd have to check
    what went into that decision - I have a vague suspicion that that behavior
    is important to incendio.

    Returns
    -------
    tuple[list]: Each item in the returned tuple is a list-like object. If the
    corresponding input arg was not like-like, it becomes the first value in
    a new list, otherwise we keep the input type (so it could be a tuple, for
    example).
    """
    res = []
    for arg in args:
        if not listlike(arg):
            arg = [arg]
        res.append(dtype(arg))
    return res


def register(name, mapping):
    """Decorator to add a function to some module-level dict mapping a string
    to a function.

    Parameters
    ----------
    name
    mapping

    Examples
    --------
    META = {}

    @register('foo', META)
    def stream_foo(x, y):
        return x * y

    >>> META
    {'foo': <function __main__.stream_foo(x, y, **kwargs)>}
    """
    def decorator(func):
        mapping[name] = func
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return decorator


def namecase(name):
    """Similar to titlecase but don't force lowercase letters after the first
    one in each word:

    name = 'Yann LeCun'
    name.title()   # 'Yann Lecun'
    namecase(name) # 'Yann LeCun'
    """
    chunks = name.split()
    return ' '.join(chunk[0].upper() + chunk[1:] for chunk in chunks)


def seconds_til_midnight(dt=None):
    """Compute seconds til midnight so we know how long to sleep for when
    changing date-based log file names. See
    jabberwocky.openai_utils.GPTBackend.update_log_path().
    """
    dt = dt or datetime.today()
    midnight = datetime.combine(dt, time())
    return (midnight - dt).seconds


"""Core functionality that ties together multiple APIs."""

from fuzzywuzzy import fuzz
import numpy as np
import pandas as pd
import re
from string import punctuation
import warnings
from youtube_transcript_api import YouTubeTranscriptApi, NoTranscriptFound

from htools import flatten, ifnone, Args, auto_repr, add_docstring, fallback
from jabberwocky.external_data import video_id
from jabberwocky.openai_utils import punctuate_mock_func, PromptManager


def realign_punctuated_text(df, text, skip_1st=0, margin=2):
    """Realign gpt3's punctuated version of a youtube transcript with the
    original timestamps.

    Parameters
    ----------
    df: pd.DataFrame
        Auto-generated transcript from YouTubeTranscriptApi. Contains 'text',
        'start', and 'duration' columns.
    text: str
        Punctuated text returned by GPT3.
    skip_1st: int
        Number of words at the start of each punctuated chunk to skip. The
        minimum number of words per row will be skip_1st+2. This helps protect
        slightly against cases where the true line is a common sequence that
        we encounter at the very start of the candidate
        Ex:
        unpunctuated: 'it is what it is'
        punctuated: 'It is what it is.'

        With skip_1st=0, we may mistakenly select "It is" from the punctuated
        candidate. With skip_1st=1, we select "It is what it is." as desired.
    margin: int
        Number of additional words to consider when constructing a punctuated
        chunk (list) from our punctuated text.

        Ex (margin=2):
        unpunctuated: ['this', 'is', 'true', 'however', 'we']
        punctuated: ['This', 'is', 'true', '-', 'however', 'we', 'still']
        Notice that adding punctuation can create new tokens, so with margin=0
        our candidate would be truncated prematurely (ending with 'however').
    """
    # Built-in str.split doesn't retain starting/trailing spaces correctly.
    # Probably would be fine but just keep this since it took a while to get
    # right and I don't want to break it.
    words = re.split(' ', text)
    rows = []
    start_i = 0
    for i, chunk in df.iterrows():
        chunk_words = re.split(' ', chunk.text)
        length = len(chunk_words)
        punct_words = words[start_i:start_i + length + margin]
        suff = ' '.join(chunk_words[-2:])
        scores = []
        bigrams = zip(punct_words[skip_1st:], punct_words[skip_1st + 1:])
        # Avoid list comp so we can exit early if we find a perfect match.
        for j, gram in enumerate(bigrams):
            score = fuzz.ratio(suff, ' '.join(gram).lower())
            if score == 100:
                argmax = j
                break
            scores.append(score)
        else:
            argmax = np.argmax(scores)
            if max(scores) < 80:
                warnings.warn(
                    'Max score < 80. Your rows may have gotten misaligned '
                    f'at row {i}: {chunk.text}'
                )
        punct_len = skip_1st + argmax + 2
        rows.append(' '.join(words[start_i:start_i + punct_len]))
        start_i += punct_len

    new_df = pd.DataFrame(rows, columns=['text'])
    return pd.concat((new_df, df.reset_index()[['start', 'duration']].copy()),
                     axis=1)


def na_index_chunks(chunk, mode='isnull', flat=False, col='text') -> list:
    """Given a chunk of a df that may contain null text rows, return a
    list of lists where each nested list contains the indices of a
    contiguous chunk of null rows.

    Parameters
    ----------
    chunk: pd.DataFrame
        DF that may have 1 or more rows where `col` is null.
    mode: str
        One of ('isnull', 'notnull') and determines which rows to retrieve.
    flat: bool
        If True, return a list of int indices. If False, return a list of lists
        where each nested list contains int indices for 1 contiguous chunk of
        non-null rows.
    col: str
        Name of column in `chunk` to check for null-ness.

    Returns
    -------
    list: See `flat` parameter documentation.
    """
    # Depending on mode, these are either nans or not nans.
    nans = chunk[getattr(chunk.text, mode)()]
    if nans.empty: return []
    last_idx = nans.index[-1]
    res = []
    curr_chunk = []
    prev = None
    for idx in nans.index:
        if prev is None or (idx == prev + 1):
            curr_chunk.append(idx)
        else:
            res.append(curr_chunk)
            curr_chunk = [idx]
        if idx == last_idx:
            res.append(curr_chunk)
        prev = idx
    return flatten(res) if flat else res


class UnpunctuatedTranscript:
    """An auto-generated transcript (probably from a youtube video). This
    provides some functionality to get punctuated or unpunctuated chunks.

    NOFIX: some methods are undocumented but this isn't used for anything
    at the moment so that's not a priority.
    """

    def __init__(self, df_gen, **kwargs):
        """
        Parameters
        ----------
        df_gen
        kwargs
        """
        self.df_gen = df_gen
        self.df_punct = self.df_gen.copy()
        self.df_punct['text'] = np.nan

        # Allow kwargs but no extra_kwargs at this point since the latter is
        # really meant to be a stopgap solution (which you can pass in when
        # calling the _punctuate_chunk method).
        self.manager = PromptManager('punctuate', verbose=False)
        self.kwargs = dict(kwargs)

    @property
    def df(self):
        return self.df_gen

    def _punctuate_chunk(self, df_chunk, extra_kwargs=None, **kwargs) -> str:
        # Don't use stream=True or return_full=True here. Just want a string.
        text = ' '.join(df_chunk.text)
        max_tokens = int(len(text.split()) * 2)
        kwargs = dict(self.kwargs, **kwargs, max_tokens=max_tokens)
        if kwargs.get('mock', False):
            kwargs['mock_func'] = punctuate_mock_func
        return self.manager.query(task='punctuate',
                                  text=text,
                                  extra_kwargs=extra_kwargs,
                                  **kwargs)[1]

    # This version only punctuates rows of the relevant chunk that haven't
    # been previously punctuated. While this is faster and cheaper, I've seen
    # some hints that the punctuation task may work better when we pass it
    # long-ish chunks of text and not little partial snippets. So it might
    # actually be better not to do this? I was going to say it might do better
    # if given full sentences rather than fragments, but I guess we can't
    # easily extract those without doing the actual punctuation step.
    def punctuated_chunk(self, start_idx, end_idx, punctuate,
                         align_kwargs=None, extra_kwargs=None,
                         **query_kwargs):
        unpunct_chunk = self.df_gen.loc[start_idx:end_idx, :]
        if not punctuate:
            return unpunct_chunk

        # If our punctuated df has any null chunks, we only want to punctuate
        # them if specifically asked to.
        chunk = self.df_punct.loc[start_idx:end_idx, :]
        if punctuate == 'if_cached' and chunk.text.isnull().sum() > 0:
            return unpunct_chunk

        # When the whole chunk is pre-punctuated, na_index_chunks returns
        # an empty list so we're not doing any unnecessary gpt3 querying.
        for idx in na_index_chunks(chunk):
            df_chunk = self.df_gen.loc[idx]
            text_punct = self._punctuate_chunk(df_chunk,
                                               extra_kwargs=extra_kwargs,
                                               **query_kwargs)
            df_chunk_punct = realign_punctuated_text(
                df_chunk, text_punct, **ifnone(align_kwargs, {})
            )
            # Use 'values' attribute because realignment func resets index.
            self.df_punct.loc[idx, 'text'] = df_chunk_punct.text.values
        return self.df_punct.loc[start_idx:end_idx, :]

    def clear_punctuations(self):
        self.df_punct['text'] = np.nan


class PunctuatedTranscript:
    """Transcript that was manually generated by a human (i.e. it has things
    like punctuation, capitalization, and is generally high quality). Like
    UnpunctuatedTranscript, this will usually come from YouTube (some popular
    videos have been transcribed by creators or fans).

    NOFIX: some methods are undocumented but this isn't used for anything
    at the moment so that's not a priority.
    """

    # Automated YouTube transcripts seem to skip all punctuation except single
    # quotes. Just delete most but replace a few with a space. This risks
    # creating multiple spaces so we also try to replace those. Save space as
    # a variable to avoid typos and make it explicit when we want to replace
    # with a space vs. nothing.
    space = ' '
    punct_rm = '|'.join(re.escape(char) for char in punctuation
                        if char not in ('/', '-'))
    punct_space = '|'.join(map(re.escape, ['/', '-', space * 2, space * 3]))

    def __init__(self, df_gen, df_punct, **kwargs):
        """
        kwargs:
            Just for compatibility with UnpunctuatedTranscript, which needs
            these to specify args like 'rstrip' when loading a prompt.
        """
        self.df_gen = df_gen
        self.df_punct = df_punct

    @property
    def df(self):
        return self.df_punct

    def punctuated_chunk(self, start_idx, end_idx, punctuate=True, **kwargs):
        chunk = self.df_punct.loc[start_idx:end_idx, :]
        # Notice this covers case where punctuate is True or a string.
        if punctuate:
            return chunk
        return chunk.assign(
            text=lambda x: x.text.str.lower()
                .str.replace(self.punct_rm, '')
                .str.replace(self.punct_space, self.space)
        )

    def clear_punctuations(self):
        warnings.warn('This is a manual transcript so there are no gpt3 '
                      'punctuations to clear.')


@auto_repr
class Transcript:

    def __init__(self, url, **kwargs):
        self.url = url
        self.id = video_id(url)
        self._transcript = self._fetch_transcripts(url, **kwargs)
        self.is_generated = isinstance(self._transcript,
                                       UnpunctuatedTranscript)
        # These are technically the start times for the first and last time
        # segments, which are slightly different from the video start and end
        # times.
        self.start_time, self.end_time = self.df.start.ends(1)

    def punctuation_kwargs(self, **kwargs):
        if self.is_generated:
            return self._transcript.manager.kwargs('punctuate', **kwargs)
        warnings.warn('No punctuation kwargs because transcipt is '
                      'pre-punctuated.')
        return {}

    def _time_range(self, start, end) -> pd.DataFrame:
        assert end > start, 'End time must be later than start time.'
        assert start >= 0 and end >= 0, 'Times must be non-negative.'
        assert start < self.end_time, \
            f'Timeframe must start before {self.end_time}.'

        df = self.df
        if start < self.start_time:
            start_idx = 0
        else:
            start_idx = df.loc[df.start <= start].index[-1]

        if end > df.start.iloc[-1]:
            end_idx = df.tail(1).index[0]
        else:
            end_idx = df.loc[df.start >= end].index[0]
        return df.iloc[start_idx:end_idx + 1]

    def time_range(self, start, end, punctuate='if_cached', align_kwargs=None,
                   extra_kwargs=None, **query_kwargs) -> pd.DataFrame:
        chunk = self._time_range(start, end)
        return self._transcript.punctuated_chunk(*chunk.ends(1).index,
                                                 punctuate=punctuate,
                                                 align_kwargs=align_kwargs,
                                                 extra_kwargs=extra_kwargs,
                                                 **query_kwargs)

    def time_range_str(self, start, end, punctuate='if_cached',
                       full_sentences=True, max_trim=120, margin=3,
                       align_kwargs=None, extra_kwargs=None,
                       **query_kwargs) -> str:
        if full_sentences and self.is_generated and punctuate is False:
            warnings.warn('This is an autogenerated transcript, so calling '
                          'time_range_str() with full_sentences=True will '
                          'not work as expected when punctuate=False. We '
                          'suggest setting these to both be True or both be '
                          'False.')

        # Values outside the acceptable range are handled later anyway. Widen
        # our candidate window so we can trim off partial sentences. May need
        # to experiment with right adjustment size.
        if full_sentences:
            start = max(0, start - margin)
            end += margin

        rows = self.time_range(start, end, punctuate=punctuate,
                               align_kwargs=align_kwargs,
                               extra_kwargs=extra_kwargs, **query_kwargs)
        text = ' '.join(rows.text.values)
        if not full_sentences:
            return text
        return self._full_sentences(text, max_trim=max_trim)

    @staticmethod
    def _full_sentences(text, max_trim=120, chars=('.', '!', '?')) -> str:
        first_upper = re.search('[A-Z]', text)
        start_idx = 0 if first_upper is None else first_upper.start()
        if start_idx > max_trim: start_idx = 0
        # Rfind returns -1 for missing chars.
        end_idx = max(text.rfind(char) for char in chars)
        if end_idx == -1 or end_idx < len(text) - max_trim - 1:
            end_idx = None
        else:
            end_idx += 1
        return text[start_idx:end_idx]

    @property
    def df(self):
        return self._transcript.df

    def _fetch_transcripts(self, url, **kwargs):
        """Wrapper to fetch youtube transcripts and create the appropriate
        transcript object depending on whether a manually generated (i.e.
        punctuated) transcript was retrieved.

        Parameters
        ----------
        url: str
        verbose: bool
        """
        df_gen, df_man, _ = self.get_transcripts(
            url, verbose=kwargs.get('verbose', True)
        )
        if df_man is None:
            return UnpunctuatedTranscript(df_gen, **kwargs)
        else:
            return PunctuatedTranscript(df_gen, df_man, **kwargs)

    def punctuated_index(self, flat=True) -> list:
        """Get indices of rows which have already been punctuated."""
        return na_index_chunks(self._transcript.df_punct, 'notnull', flat)

    def unpunctuated_index(self, flat=True) -> list:
        """Get indices of rows which have not yet been punctuated."""
        return na_index_chunks(self._transcript.df_punct, 'isnull', flat)

    def punctuated_times(self):
        df = self.df
        res = []
        for chunk in self.punctuated_index(flat=False):
            end_row = df.loc[chunk[-1]]
            res.append((df.loc[chunk[0], 'start'],
                        end_row.start + end_row.duration))
        return res

    def punctuated_time_rows(self, chunk=False):
        idx = self.punctuated_index(flat=not chunk)
        # Don't use self.df, that points to unpunctuated version for generated
        # transcript.
        df = self._transcript.df_punct
        if chunk: return [df.loc[i] for i in idx]
        return df.loc[idx]

    def clear_punctuations(self):
        # Useful if we've been testing with mock calls and want to reset.
        self._transcript.clear_punctuations()

    @staticmethod
    def get_transcripts(url, verbose=True):
        """Fetch one or more transcripts for a youtube video given its URL.

        Parameters
        ----------
        url: str
            Don't include any channel-related suffix. E.G. use
            https://www.youtube.com/watch?v=OZbCRN3C_Hs, not
            https://www.youtube.com/watch?v=OZbCRN3C_Hs&ab_channel=BBC.
        verbose: bool
            Warn

        Returns
        -------
        DotDict: Contains keys 'id' (maps to video ID str), 'generated',
        and 'manual' (the latter two lap to pandas dfs or None if no
        transcript was found). Manual transcripts are human-created.
        Generated transcripts are a bit lower quality and tend to lack
        punctuation.
        """
        langs = ['en', 'en-GB']
        id_ = video_id(url)
        res = {'generated': None, 'manual': None}
        trans_list = YouTubeTranscriptApi.list_transcripts(id_)
        res['generated'] = trans_list.find_generated_transcript(langs)
        try:
            res['manual'] = trans_list.find_manually_created_transcript(langs)
        except NoTranscriptFound:
            if verbose: warnings.warn('No manual transcript found.')
        if verbose:
            non_eng = [k for k, v in res.items()
                       if v and ('United Kingdom' in v.language)]
            if non_eng:
                warnings.warn(
                    f'{non_eng} {"has" if len(non_eng) == 1 else "have"} '
                    'language en-GB, not en.'
                )
        return Args(**{k: pd.DataFrame(v.fetch()) if v else v
                       for k, v in res.items()},
                    id=id_)

    def __str__(self):
        return f'{type(self).__name__}(url={self.url}, ' \
               f'is_generated={self.is_generated})'


class Session:
    """Manage multiple transcripts."""

    def __init__(self, *tasks, **video_urls):
        self.transcripts = {name: Transcript(url)
                            for name, url in video_urls.items()}
        self.manager = PromptManager(*tasks, verbose=False)

    @add_docstring(Transcript.time_range_str)
    def query(self, video, task, start_time, end_time, punctuate=True,
              punct_extra_kwargs=None, punct_query_kwargs=None,
              task_extra_kwargs=None, **task_kwargs):
        text = self[video].time_range_str(
            start_time, end_time,
            punctuate=punctuate,
            extra_kwargs=punct_extra_kwargs or {},
            **ifnone(punct_query_kwargs, {})
        )
        prompt, resp = self.manager.query(task, text,
                                          extra_kwargs=task_extra_kwargs,
                                          **task_kwargs)
        return prompt, resp

    def __getitem__(self, key):
        return self.transcripts[key]

    def __iter__(self):
        return iter(self.transcripts)

    def __repr__(self):
        name = type(self).__name__
        sep = ',\n' + ' ' * (len(name) + 1)
        kwarg_strs = sep.join(f'{name}={trans.url!r}' for name, trans
                              in self.transcripts.items())
        arg_strs = sep.join(repr(p) for p in self.manager.prompts)
        return f'{name}({arg_strs}{sep}{kwarg_strs})'


class GuiTextChunker:
    """Used by GUI to display text with a max line length. Dearpygui doesn't
    have any auto-text wrap that I could find (at least in this version) so
    longer responses would otherwise trail way past the defined text box. I
    don't like the huge horizontal scrollbars that result so this is my
    workaround.

    This maintains both "chunked" (max line length enforced) and non-chunked
    versions of an input text because we often prefer not to send GPT-3 text
    with newlines haphazardly inserted. I.e. we show the user the chunked
    version but query with the raw version.
    """

    def __init__(self, max_chars=79):
        self.raw = {}
        self.chunked = {}
        self.max_chars = max_chars

        # Transcriber and gpt3 generally don't insert \n\r or \r\n so this is
        # a decent # way to recover where I inserted splits. Otherwise, we risk
        # accumulating more and more newlines when calling this repeatedly.
        # Still feel like the check for previously added text in self.add()
        # should avoid this but sometimes the prompt in GUI still shows
        # text with what appear to be extra newlines.x
        self.newline = '\r\n'
        self.newline_tmp = '<NEWLINE>'

    @fallback(keep=['max_chars'])
    def add(self, key, text, return_chunked=True, sanitize_for_gui=True,
            **kwargs):
        if sanitize_for_gui: text = self.gui_sanitize(text)
        if self._previously_added(key, text):
            if return_chunked:
                return self.get(key, chunked=True)
            return
        # Want to be extra sure we don't re-chunk as this can cause bugs.
        text = self.to_raw(text)
        chunked = self._chunk_lines(text, max_chars)
        self.raw[key] = text
        self.chunked[key] = chunked
        if return_chunked: return chunked

    def get(self, key, chunked):
        if chunked:
            return self.chunked[key]
        return self.raw[key]

    def delete(self, key):
        del self.raw[key]
        del self.chunked[key]

    @staticmethod
    def sticky_split(text, sep):
        """Basically a functioning version of htools.hsplit with group=True and
        attach=True. Realized that doesn't work when sep contains special
        characters.
        """
        res = []
        toks = text.split(sep)
        max_idx = len(toks) - 1
        for i, tok in enumerate(toks):
            if tok:
                if i < max_idx: tok += sep
                res.append(tok)
            elif res and i < max_idx:
                res[-1] += sep
        return res

    def _chunk_lines(self, text, max_chars):
        text = text.replace('\n', self.newline_tmp).rstrip(self.newline_tmp)
        words = [word for row in self.sticky_split(text, self.newline_tmp)
                 for word in row.split(' ')]

        lines, line = [], []
        curr_len = 0
        for word in words:
            length = len(word) + 1
            prev_has_newline = line and self.newline_tmp in line[-1]
            if (curr_len + length > max_chars) or prev_has_newline:
                if not prev_has_newline:
                    line[-1] = line[-1] + self.newline
                else:
                    line[-1] = line[-1].replace(self.newline_tmp, '\n')
                lines.append(line)
                line = []
                curr_len = 0
            line.append(word)
            curr_len += length
        if line:
            lines.append([word.replace(self.newline_tmp, '\n')
                          for word in line])
        return ''.join(' '.join(line) for line in lines)

    def to_raw(self, text):
        return text .replace(self.newline, ' ') \
            .replace(self.newline_tmp, '\n') \
            .rstrip('\n')

    def _previously_added(self, key, text, sanitize_for_gui=True):
        try:
            raw = self.get(key, chunked=False)
            # A bit redundant since this is mostly used in add() which already
            # does this, but want to protect against case where someone uses
            # this method by itself (which I've already done - maybe a sign
            # that is shouldn't be internal).
            if sanitize_for_gui: text = self.gui_sanitize(text)
            assert self.to_raw(text) == self.to_raw(raw)
            return True
        except (KeyError, AssertionError) as e:
            print(e)
            return False

    def clear(self):
        self.raw.clear()
        self.chunked.clear()

    def gui_sanitize(self, text):
        """Replace characters that gpt3 often uses but dearpygui can't render.
        First dash we replace is en dash, second is em dash.
        """
        return text.replace('‘', "'").replace('’', "'")\
                   .replace('“', '"').replace('”', '"')\
                   .replace('–', '-').replace('—', '-')

    def __contains__(self, key):
        in_raw, in_chunked = key in self.raw, key in self.chunked
        if in_raw and in_chunked:
            return True
        elif in_raw or in_chunked:
            raise KeyError(
                f'Key {key} was found in '
                f'{"self.raw" if in_raw else "self.chunked"}. Should be in '
                'neither or both. It may be wise to call the clear() method '
                'and re-add your key.'
            )
        else:
            return False



import os
from pathlib import Path
import setuptools


def requirements(path='requirements.txt'):
    with open(path, 'r') as f:
        deps = [line.strip() for line in f]
    return deps


def version():
    path = os.path.join('jabberwocky', '__init__.py')
    with open(path, 'r') as f:
        for row in f:
            if not row.startswith('__version__'):
                continue
            return row.split(' = ')[-1].strip('\n').strip("'")


def file_text(fname):
    path = Path(__file__).parent/fname
    return path.read_text()


setuptools.setup(name='jabberwocky',
                 version=version(),
                 author='Harrison Mamin',
                 author_email='harrisonmamin@gmail.com',
                 description='Core library powering a GUI providing an audio '
                             'interface to GPT3.',
                 install_requires=requirements(),
                 packages=setuptools.find_packages(),
                 long_description=file_text('README.md'),
                 long_description_content_type='text/markdown')

